"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class DeviceCreated(google.protobuf.message.Message):
    """error: no message type"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FOOBAR_FIELD_NUMBER: builtins.int
    foobar: builtins.int = ...
    def __init__(self,
        *,
        foobar : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> None: ...
global___DeviceCreated = DeviceCreated

class TripCreatedV3(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    DEVICE_ID_FIELD_NUMBER: builtins.int
    device_id: typing.Text = ...
    def __init__(self,
        *,
        device_id : typing.Optional[typing.Text] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["device_id",b"device_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["device_id",b"device_id"]) -> None: ...
global___TripCreatedV3 = TripCreatedV3

class TripCreated(google.protobuf.message.Message):
    """error: duplicate"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FOOBAR_FIELD_NUMBER: builtins.int
    foobar: builtins.int = ...
    def __init__(self,
        *,
        foobar : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> None: ...
global___TripCreated = TripCreated

class TripCreatedV4(google.protobuf.message.Message):
    """error: no message options"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FOOBAR_FIELD_NUMBER: builtins.int
    foobar: builtins.int = ...
    def __init__(self,
        *,
        foobar : typing.Optional[builtins.int] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["foobar",b"foobar"]) -> None: ...
global___TripCreatedV4 = TripCreatedV4



from hedwig.testing.factories.protobuf import ProtobufMessageFactory as BaseProtobufMessageFactory
from tests.schemas.protos import protobuf_pb2


class ProtobufMessageFactory(BaseProtobufMessageFactory):
    class Params:
        protobuf_schema_module = protobuf_pb2



from distutils.version import StrictVersion

import pytest


@pytest.fixture(autouse=True, scope="module")
def setup_instrumentation():
    from hedwig.instrumentation.compat import set_global_textmap
    import opentelemetry.trace
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.trace import set_tracer_provider
    from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
    from opentelemetry.version import __version__

    VERSION_1_6_0 = StrictVersion("1.6.0")

    try:
        # might already be set to default tracer provider
        opentelemetry.trace._TRACER_PROVIDER = None
        if StrictVersion(__version__) >= VERSION_1_6_0:
            opentelemetry.trace._TRACER_PROVIDER_SET_ONCE._done = False

        set_tracer_provider(TracerProvider())

        set_global_textmap(TraceContextTextMapPropagator())

        yield
    finally:
        opentelemetry.trace._TRACER_PROVIDER = None
        if StrictVersion(__version__) >= VERSION_1_6_0:
            opentelemetry.trace._TRACER_PROVIDER_SET_ONCE._done = False



from unittest import mock

import pytest

get_tracer = pytest.importorskip('opentelemetry.trace.get_tracer')


@mock.patch('hedwig.backends.base.Message.exec_callback', autospec=True)
def test_message_handler_updates_span_name(mock_exec_callback, message, consumer_backend):
    provider_metadata = mock.Mock()
    tracer = get_tracer(__name__)
    with tracer.start_as_current_span(test_message_handler_updates_span_name.__name__, {}) as span:
        assert span.name == test_message_handler_updates_span_name.__name__
        consumer_backend.message_handler(*message.serialize(), provider_metadata)
        assert span.name == message.type
        assert span.get_span_context().is_valid



import logging
from typing import Optional, Dict, Any

try:
    import structlog
except ImportError:  # pragma: no cover
    structlog = None  # type: ignore


def log(module: str, level: int, message: str, exc_info: Optional[bool] = None, extra: Optional[Dict[Any, Any]] = None):
    kwargs: Dict[Any, Any] = {}
    if exc_info is not None:
        kwargs["exc_info"] = True
    if structlog:
        method_name = logging.getLevelName(level).lower()
        if extra:
            kwargs.update(extra)
        logger = structlog.getLogger(module)
        method = getattr(logger, method_name)
        method(message, **kwargs)
    else:
        if extra:
            kwargs["extra"] = extra
        logging.getLogger(module).log(level, message, **kwargs)



#!/usr/bin/env python
import logging
import os
import importlib
import typing

from hedwig.backends.import_utils import import_module_attr
from hedwig.utils import log

try:
    from django.conf import settings as django_settings
    from django.dispatch import receiver
    from django.test import signals

    HAVE_DJANGO = True  # pragma: no cover
except ImportError:
    HAVE_DJANGO = False

try:
    from flask import current_app  # pragma: no cover

    HAVE_FLASK = True  # pragma: no cover
except ImportError:
    HAVE_FLASK = False

_DEFAULTS: dict = {
    'AWS_REGION': None,
    'AWS_ACCOUNT_ID': None,
    'AWS_ACCESS_KEY': None,
    'AWS_CONNECT_TIMEOUT_S': 2,
    'AWS_ENDPOINT_SNS': None,
    'AWS_ENDPOINT_SQS': None,
    'AWS_READ_TIMEOUT_S': 2,
    'AWS_SECRET_KEY': None,
    'AWS_SESSION_TOKEN': None,
    'GOOGLE_APPLICATION_CREDENTIALS': None,
    'GOOGLE_CLOUD_PROJECT': None,
    'GOOGLE_PUBSUB_READ_TIMEOUT_S': 5,
    'HEDWIG_CALLBACKS': {},
    'HEDWIG_CONSUMER_BACKEND': None,
    'HEDWIG_DATA_VALIDATOR_CLASS': 'hedwig.validators.jsonschema.JSONSchemaValidator',
    'HEDWIG_DEFAULT_HEADERS': 'hedwig.conf.default_headers_hook',
    'HEDWIG_HEARTBEAT_INACTIVITY_RESET_S': None,
    'HEDWIG_HEARTBEAT_INTERVAL_S': 15,
    'HEDWIG_HEARTBEAT_HOOK': 'hedwig.conf.noop_hook',
    'HEDWIG_MESSAGE_ROUTING': {},
    'HEDWIG_PRE_PROCESS_HOOK': 'hedwig.conf.noop_hook',
    'HEDWIG_POST_PROCESS_HOOK': 'hedwig.conf.noop_hook',
    'HEDWIG_PUBLISHER': None,
    'HEDWIG_PUBLISHER_BACKEND': None,
    'HEDWIG_PUBLISHER_GCP_BATCH_SETTINGS': (),
    'HEDWIG_QUEUE': None,
    'HEDWIG_JSONSCHEMA_FILE': None,
    'HEDWIG_PROTOBUF_MESSAGES': None,
    'HEDWIG_SYNC': False,
    'HEDWIG_SUBSCRIPTIONS': [],
    'HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES': True,
}


# List of settings that may be in string import notation.
_IMPORT_STRINGS = (
    'HEDWIG_CONSUMER_BACKEND',
    'HEDWIG_DATA_VALIDATOR_CLASS',
    'HEDWIG_DEFAULT_HEADERS',
    'HEDWIG_HEARTBEAT_HOOK',
    'HEDWIG_PRE_PROCESS_HOOK',
    'HEDWIG_POST_PROCESS_HOOK',
    'HEDWIG_PUBLISHER_BACKEND',
)

# List of settings that will be dicts with values as string import notation.
_IMPORT_DICT_VALUES = ('HEDWIG_CALLBACKS',)

# List of settings that will be lists with values as string import notation.
_IMPORT_LIST_VALUES = ('HEDWIG_PROTOBUF_MESSAGES',)


def default_headers_hook(*args, **kwargs) -> typing.Dict[str, str]:
    return {}


def noop_hook(*args, **kwargs) -> None:
    pass


class _LazySettings:
    """
    A lazy object, that allows settings to be accessed as properties.
    For example:

    .. code-block:: python

        from hedwig.conf import settings
        print(settings.AWS_REGION)

    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self) -> None:
        self._defaults = _DEFAULTS
        self._import_strings = _IMPORT_STRINGS
        self._import_dict_values = _IMPORT_DICT_VALUES
        self._import_list_values = _IMPORT_LIST_VALUES
        self._user_settings: object = None

    @property
    def configured(self) -> bool:
        """
        Have Hedwig settings been configured?
        """
        return bool(self._user_settings)

    def _ensure_configured(self):
        if self.configured:
            return

        if os.environ.get("SETTINGS_MODULE"):
            log(__name__, logging.INFO, f'Configuring Hedwig through module: {os.environ["SETTINGS_MODULE"]}')
            self._user_settings = importlib.import_module(os.environ["SETTINGS_MODULE"], package=None)
        elif HAVE_DJANGO:  # pragma: no cover
            log(__name__, logging.INFO, 'Configuring Hedwig through django settings')
            # automatically import Django settings in Django projects
            self._user_settings = django_settings
        elif HAVE_FLASK:  # pragma: no cover
            log(__name__, logging.INFO, 'Configuring Hedwig through flask settings')
            # automatically import Flask settings in Flask projects
            self._user_settings = current_app.config
        if not self._user_settings:
            raise ImportError("Hedwig settings have not been configured")

    def configure_with_object(self, obj: object) -> None:  # pragma: no cover
        """
        Set Hedwig config using a dataclass-like object that contains all settings as its attributes, or a dict that
        contains settings as its keys.
        """
        assert not self._user_settings, "Hedwig settings have already been configured"

        log(__name__, logging.INFO, 'Configuring Hedwig through object')
        self._user_settings = obj

    @staticmethod
    def _import_string(
        dotted_path_or_callable: typing.Union[typing.Callable, str]
    ) -> typing.Union[typing.Callable, typing.Type]:
        """
        Import a dotted module path and return the attribute/class designated by the
        last name in the path. Raise ImportError if the import failed.
        """
        if callable(dotted_path_or_callable):
            return dotted_path_or_callable

        return import_module_attr(dotted_path_or_callable)

    def _get_setting_from_object(self, attr: str):
        if isinstance(self._user_settings, dict):  # pragma: no cover
            if attr in self._user_settings:
                return self._user_settings[attr]
            elif attr.lower() in self._user_settings:
                return self._user_settings[attr.lower()]
            raise RuntimeError
        else:
            try:
                # Check if present in user settings
                return getattr(self._user_settings, attr)
            except AttributeError:
                # try lowercase
                try:
                    return getattr(self._user_settings, attr.lower())
                except AttributeError:
                    raise RuntimeError

    def __getattr__(self, attr: str) -> typing.Any:
        self._ensure_configured()

        if attr not in self._defaults:
            raise AttributeError("Invalid setting: '%s'" % attr)

        try:
            val = self._get_setting_from_object(attr)
        except RuntimeError:
            # Fall back to defaults
            val = self._defaults[attr]

        # Coerce import strings into classes
        if attr in self._import_strings:
            val = self._import_string(val)

        if attr in self._import_dict_values:
            val = {k: self._import_string(v) for k, v in val.items()}

        if attr in self._import_list_values:
            val = [self._import_string(v) for v in val]

        # Cache the result
        setattr(self, attr, val)
        return val

    def clear_cache(self) -> None:
        """
        Clear settings cache - useful for testing only
        """
        from hedwig.backends.utils import get_publisher_backend, get_consumer_backend
        from hedwig.callback import Callback
        from hedwig.models import _validator

        for attr in self._defaults:
            try:
                delattr(self, attr)
            except AttributeError:
                pass
        Callback.find_by_message.cache_clear()

        # since consumer/publisher settings may have changed
        get_publisher_backend.cache_clear()
        get_consumer_backend.cache_clear()

        # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
        _validator.cache_clear()


if HAVE_DJANGO:  # pragma: no cover

    @receiver(signals.setting_changed)
    def clear_cache_on_setting_override(sender, setting, value, enter, **kwargs):
        settings.clear_cache()


settings = _LazySettings()
"""
This object allows settings to be accessed as properties. Settings can be configured in one of three ways:

#. Environment variable named ``SETTINGS_MODULE`` that points to a python module with settings as module attributes

#. Django - if Django can be imported, Django settings will be used automatically

#. Using an object or dict, by calling :meth:`hedwig.conf.settings.configure_with_object`

Some setting values need to be string import paths will be automatically resolved and return the class.

Once settings have been configured, they shouldn't be changed. It is possible to re-configure for testing, but its not
guaranteed to work the same way for non-test use cases.
"""



import os


# to test callables in HEDWIG_CALLBACKS
def device_handler(message):
    pass


AWS_ACCESS_KEY = "DUMMY_KEY"
AWS_ACCOUNT_ID = "DUMMY_ACCOUNT"
AWS_REGION = "DUMMY_REGION"
AWS_SECRET_KEY = "DUMMY_SECRET"

AWS_CONNECT_TIMEOUT_S = 5
AWS_READ_TIMEOUT_S = 5

HEDWIG_QUEUE = 'DEV-MYAPP'

HEDWIG_JSONSCHEMA_FILE = os.path.abspath('tests/schemas/jsonschema.json')

HEDWIG_PROTOBUF_MESSAGES = [
    "tests.schemas.protos.protobuf_pb2.TripCreatedV1",
    "tests.schemas.protos.protobuf_pb2.TripCreatedV2",
    "tests.schemas.protos.protobuf_pb2.DeviceCreatedV1",
    "tests.schemas.protos.protobuf_pb2.VehicleCreatedV1",
]

HEDWIG_CALLBACKS = {
    ('trip_created', '1.*'): 'tests.handlers.trip_created_handler',
    ('trip_created', '2.*'): 'tests.handlers.trip_created_handler',
    ('device.created', '1.*'): device_handler,
}

HEDWIG_MESSAGE_ROUTING = {
    ('trip_created', '1.*'): 'dev-trip-created-v1',
    ('trip_created', '2.*'): 'dev-trip-created-v2',
    ('device.created', '1.*'): 'dev-device-created-v1',
    ('vehicle_created', '1.*'): ('dev-vehicle-created-v1', 'project-id-or-account-id'),
}

HEDWIG_PUBLISHER = 'myapp'



import base64
from typing import List, Type

import pytest

from hedwig.exceptions import ValidationError

pytest.importorskip('google.protobuf')

from google.protobuf import json_format  # noqa
from google.protobuf.struct_pb2 import Value  # noqa
from google.protobuf.message import Message as ProtoMessage  # noqa

from hedwig.testing.factories.protobuf import ProtobufMessageFactory  # noqa
from hedwig.validators.protobuf import ProtobufValidator, SchemaError  # noqa
from hedwig.protobuf.container_pb2 import PayloadV1  # noqa
from tests.models import MessageType  # noqa
from tests.schemas.protos import (  # noqa
    protobuf_pb2,
    protobuf_minor_versioned_pb2,
    protobuf_bad_pb2,
)
from tests import test_validators  # noqa


class TestProtobufValidator:
    def _validator(self):
        return ProtobufValidator()

    @pytest.mark.parametrize(
        'proto_messages,schema_exc_error',
        [
            [
                [
                    protobuf_bad_pb2.DeviceCreated,
                    protobuf_pb2.TripCreatedV1,
                    protobuf_pb2.TripCreatedV2,
                    protobuf_pb2.DeviceCreatedV1,
                    protobuf_pb2.VehicleCreatedV1,
                ],
                "Couldn't determine message type for Protobuf message class '<class 'protobuf_bad_pb2.DeviceCreated'>'",
            ],
            [
                [
                    protobuf_pb2.TripCreatedV1,
                    protobuf_pb2.DeviceCreatedV1,
                    protobuf_pb2.VehicleCreatedV1,
                ],
                "Invalid message_type, major version: 'trip_created/2', not found in declared messages",
            ],
            [
                [
                    protobuf_bad_pb2.TripCreatedV3,
                    protobuf_pb2.TripCreatedV1,
                    protobuf_pb2.TripCreatedV2,
                    protobuf_pb2.DeviceCreatedV1,
                    protobuf_pb2.VehicleCreatedV1,
                ],
                "Protobuf message class '<class 'protobuf_bad_pb2.TripCreatedV3'>' mismatch in major version: '3' !="
                " '2'",
            ],
            [
                [
                    protobuf_bad_pb2.TripCreated,
                    protobuf_pb2.TripCreatedV1,
                    protobuf_pb2.TripCreatedV2,
                    protobuf_pb2.DeviceCreatedV1,
                    protobuf_pb2.VehicleCreatedV1,
                ],
                "Duplicate Protobuf message declared for 'trip_created/1'",
            ],
            [
                [
                    protobuf_bad_pb2.TripCreatedV4,
                    protobuf_pb2.TripCreatedV1,
                    protobuf_pb2.TripCreatedV2,
                    protobuf_pb2.DeviceCreatedV1,
                    protobuf_pb2.VehicleCreatedV1,
                ],
                "Protobuf message class '<class 'protobuf_bad_pb2.TripCreatedV4'>' does not define option "
                "message_options",
            ],
        ],
    )
    def test_check_schema(self, proto_messages: List[Type[ProtoMessage]], schema_exc_error):
        with pytest.raises(SchemaError) as exc_context:
            ProtobufValidator(proto_messages)

        assert schema_exc_error in exc_context.value.args[0]

    def test_serialize(self, use_transport_message_attrs):
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created, model_version=1, protobuf_schema_module=protobuf_pb2
        )
        if not use_transport_message_attrs:
            msg = PayloadV1()
            msg.format_version = '1.0'
            msg.schema = 'trip_created/1.0'
            msg.id = message.id
            msg.metadata.timestamp.FromMilliseconds(message.timestamp)
            msg.metadata.publisher = message.publisher
            for k, v in message.headers.items():
                msg.metadata.headers[k] = v
            msg.data.Pack(message.data)
            attributes = message.headers
            serialized = self._validator().serialize(message)
            serialized_msg = PayloadV1()
            serialized_msg.ParseFromString(serialized[0])
            assert msg == serialized_msg
            assert attributes == serialized[1]
        else:
            msg = message.data
            attributes = {
                "hedwig_format_version": '1.0',
                "hedwig_schema": "trip_created/1.0",
                "hedwig_id": message.id,
                "hedwig_publisher": message.publisher,
                "hedwig_message_timestamp": str(message.timestamp),
                **message.headers,
            }
            serialized = self._validator().serialize(message)
            serialized_msg = protobuf_pb2.TripCreatedV1()
            serialized_msg.ParseFromString(serialized[0])
            assert attributes == serialized[1]
            assert msg == serialized_msg

    @pytest.mark.parametrize("unknown_schema", [True, False])
    def test_serialize_firehose(self, unknown_schema, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect firehose serialization
        _ = use_transport_message_attrs
        minor_version = 1 if unknown_schema else 0
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created,
            model_version=1,
            addition_version=minor_version,
            protobuf_schema_module=protobuf_minor_versioned_pb2 if unknown_schema else protobuf_pb2,
        )
        msg = PayloadV1()
        msg.format_version = '1.0'
        msg.schema = f'trip_created/1.{minor_version}'
        msg.id = message.id
        msg.metadata.timestamp.FromMilliseconds(message.timestamp)
        msg.metadata.publisher = message.publisher
        for k, v in message.headers.items():
            msg.metadata.headers[k] = v
        if unknown_schema:
            value_msg = Value()
            value_msg.string_value = base64.b64encode(message.data.SerializeToString())
            msg.data.Pack(value_msg)
        else:
            msg.data.Pack(message.data)
        serialized = self._validator().serialize_firehose(message)
        serialized_msg = json_format.Parse(serialized, PayloadV1())
        assert msg == serialized_msg

    def test_serialize_containerized(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect containerized serialization
        _ = use_transport_message_attrs
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created,
            model_version=1,
            protobuf_schema_module=protobuf_pb2,
        )
        msg = PayloadV1()
        msg.format_version = '1.0'
        msg.schema = 'trip_created/1.0'
        msg.id = message.id
        msg.metadata.timestamp.FromMilliseconds(message.timestamp)
        msg.metadata.publisher = message.publisher
        for k, v in message.headers.items():
            msg.metadata.headers[k] = v
        msg.data.Pack(message.data)
        serialized = self._validator().serialize_containerized(message)
        serialized_msg = PayloadV1()
        serialized_msg.ParseFromString(serialized)
        assert msg == serialized_msg

    def test_deserialize(self, use_transport_message_attrs):
        provider_metadata = object()
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created, model_version=1, protobuf_schema_module=protobuf_pb2
        )
        message = message.with_provider_metadata(provider_metadata)

        if not use_transport_message_attrs:
            msg = PayloadV1()
            msg.format_version = '1.0'
            msg.schema = 'trip_created/1.0'
            msg.id = message.id
            msg.metadata.timestamp.FromMilliseconds(message.timestamp)
            msg.metadata.publisher = message.publisher
            for k, v in message.headers.items():
                msg.metadata.headers[k] = v
            msg.data.Pack(message.data)
            message_payload = msg.SerializeToString()
            attributes = message.headers
        else:
            message_payload = message.data.SerializeToString()
            attributes = {
                "hedwig_format_version": '1.0',
                "hedwig_schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip_created/1.0",
                "hedwig_id": message.id,
                "hedwig_publisher": message.publisher,
                "hedwig_message_timestamp": str(message.timestamp),
                **message.headers,
            }

        assert message == self._validator().deserialize(message_payload, attributes, provider_metadata)

    @pytest.mark.parametrize("unknown_schema", [True, False])
    def test_deserialize_firehose(self, unknown_schema, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect firehose deserialization
        _ = use_transport_message_attrs
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created, model_version=1, protobuf_schema_module=protobuf_pb2
        )

        msg = PayloadV1()
        msg.format_version = '1.0'
        msg.schema = 'trip_created/1.0'
        msg.id = message.id
        msg.metadata.timestamp.FromMilliseconds(message.timestamp)
        msg.metadata.publisher = message.publisher
        for k, v in message.headers.items():
            msg.metadata.headers[k] = v
        if unknown_schema:
            value_msg = Value()
            value_msg.string_value = base64.b64encode(message.data.SerializeToString()).decode()
            msg.data.Pack(value_msg)
        else:
            msg.data.Pack(message.data)
        message_payload = json_format.MessageToJson(msg, preserving_proto_field_name=True, indent=0).replace("\n", "")

        assert message == self._validator().deserialize_firehose(message_payload)

    def test_deserialize_containerized(self, use_transport_message_attrs):
        # use_transport_message_attrs shouldn't affect containerized deserialization
        _ = use_transport_message_attrs
        message = ProtobufMessageFactory(
            msg_type=MessageType.trip_created, model_version=1, protobuf_schema_module=protobuf_pb2
        )

        msg = PayloadV1()
        msg.format_version = '1.0'
        msg.schema = 'trip_created/1.0'
        msg.id = message.id
        msg.metadata.timestamp.FromMilliseconds(message.timestamp)
        msg.metadata.publisher = message.publisher
        for k, v in message.headers.items():
            msg.metadata.headers[k] = v
        msg.data.Pack(message.data)
        message_payload = msg.SerializeToString()

        assert message == self._validator().deserialize_containerized(message_payload)

    def test_deserialize_raises_error_invalid_schema(self):
        validator = self._validator()

        payload = b''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert 'Invalid message attribute: hedwig_format_version must be string, found: None' in str(e.value.args[0])

        payload = b'{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "trip_created",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, attrs, None)
        assert e.value.args[0] == 'Invalid schema found: trip_created'

        payload = b'{}'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "trip_created/9.0",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, attrs, None)
        assert e.value.args[0] == "Protobuf message class not found for 'trip_created/9'"

    def test_deserialize_raises_error_invalid_schema_container(self, settings):
        settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES = False

        validator = self._validator()

        payload = b''
        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert '' in str(e.value.args[0])

        msg = PayloadV1()
        msg.format_version = '1.0'
        msg.schema = 'trip_created'
        msg.id = "2acd99ec-47ac-3232-a7f3-6049146aad15"
        msg.metadata.timestamp.FromMilliseconds(1)
        msg.metadata.publisher = ""
        payload = msg.SerializeToString()

        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert e.value.args[0] == 'Invalid schema found: trip_created'

        msg.schema = 'trip_created/9.0'
        payload = msg.SerializeToString()

        with pytest.raises(ValidationError) as e:
            validator.deserialize(payload, {}, None)
        assert e.value.args[0] == "Protobuf message class not found for 'trip_created/9'"

    def test_deserialize_fails_on_invalid_payload(self):
        with pytest.raises(ValidationError):
            self._validator().deserialize(b"bad proto", {}, None)

    def test_deserialize_raises_error_invalid_data(self):
        payload = b'\xaf\x95\x92x`'
        attrs = {
            "hedwig_format_version": "1.0",
            "hedwig_schema": "trip_created/1.0",
            "hedwig_id": "2acd99ec-47ac-3232-a7f3-6049146aad15",
            "hedwig_publisher": "",
            "hedwig_headers": "{}",
            "hedwig_message_timestamp": "1",
        }
        with pytest.raises(ValidationError) as e:
            self._validator().deserialize(payload, attrs, None)
        assert e.value.args[0] in [
            'Invalid data for message: TripCreatedV1: Error parsing message',
            "Invalid data for message: TripCreatedV1: Error parsing message with type 'tests.TripCreatedV1'",
        ]

    def test_serialize_no_error_invalid_data(self):
        message = ProtobufMessageFactory(
            msg_type='device.created',
            data=protobuf_bad_pb2.DeviceCreated(foobar=1),
            protobuf_schema_module=protobuf_pb2,
        )
        # invalid data is ignored so long as its a valid protobuf class
        self._validator().serialize(message)

    def test_serialize_raises_error_invalid_message_type(self):
        message = ProtobufMessageFactory(
            msg_type='invalid',
            data=protobuf_pb2.DeviceCreatedV1(device_id="abcd", user_id="U_123"),
            protobuf_schema_module=protobuf_pb2,
        )
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0] == "Protobuf message class not found for 'invalid/1'"

    def test_serialize_raises_error_invalid_minor_version(self):
        message = ProtobufMessageFactory(
            msg_type='device.created',
            addition_version=1,
            data=protobuf_pb2.DeviceCreatedV1(device_id="abcd", user_id="U_123"),
            protobuf_schema_module=protobuf_pb2,
        )
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0] == 'Unknown minor version: 1, last known minor version: 0'

    def test_serialize_raises_error_invalid_headers(self):
        message = ProtobufMessageFactory(
            msg_type='device.created',
            protobuf_schema_module=protobuf_pb2,
            metadata__headers__hedwig_foo="bar",
        )
        with pytest.raises(ValidationError) as e:
            self._validator().serialize(message)
        assert e.value.args[0] == "Invalid header key: 'hedwig_foo' - can't begin with reserved namespace 'hedwig_'"



import json
import re
import typing
from copy import deepcopy
from decimal import Decimal
from distutils.version import StrictVersion
from functools import lru_cache
from pathlib import Path
from typing import Tuple, Union, Optional
from uuid import UUID

import funcy
from funcy import cached_property
from jsonschema import SchemaError, RefResolutionError, FormatChecker
from jsonschema.validators import Draft4Validator

from hedwig.conf import settings
from hedwig.exceptions import ValidationError
from hedwig.validators.base import HedwigBaseValidator, MetaAttributes


def _json_default(obj):
    if isinstance(obj, Decimal):
        int_val = int(obj)
        if int_val == obj:
            return int_val
        else:
            return float(obj)
    elif isinstance(obj, UUID):
        return str(obj)
    raise TypeError


class JSONSchemaValidator(HedwigBaseValidator):
    checker = FormatChecker()
    """
    FormatChecker that checks for `format` JSON-schema field. This may be customized by an app by overriding setting
    `HEDWIG_DATA_VALIDATOR_CLASS` and defining more format checkers.
    """

    schema: dict
    """
    The schema to validate data against - supplied by app
    """

    # uuid separated by hyphens:
    _human_uuid_re = re.compile(r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$")

    _version_pattern_re = re.compile(r"^([0-9]+)\.\*$")

    _validator: Draft4Validator

    _container_validator: Draft4Validator

    FORMAT_VERSIONS = [StrictVersion('1.0')]
    '''
    Here are the schema definitions:

    Version 1.0:
    {
        "format_version": "1.0",
        "schema": "https://github.com/cloudchacho/hedwig-python/schema#/schemas/trip.created/1.0",
        "id": "b1328174-a21c-43d3-b303-964dfcc76efc",
        "metadata": {
            "timestamp": 1460868253255,
            "publisher": "myapp",
            "headers": {
                ...
            }
        },
        "data": {
            ...
        }
    }

    All the top-level fields (other than `metadata`) are required to be non-empty. `metadata` field is expected to
    be present, but may be empty. All fields in `metadata` are optional. `data` is validated using `schema`.
    '''

    def __init__(self, schema: typing.Optional[dict] = None) -> None:
        # automatically load schema
        container_schema_filepath = Path(__file__).resolve().parent / 'jsonschema_container_schema.json'
        with open(container_schema_filepath) as f:
            container_schema = json.load(f)

        self._container_validator = Draft4Validator(container_schema)

        if schema is None:
            # automatically load schema
            schema_filepath = settings.HEDWIG_JSONSCHEMA_FILE
            with open(schema_filepath) as f:
                schema = json.load(f)

        self._check_schema(schema)

        self.schema = schema

        self._validator = Draft4Validator(schema, format_checker=self.checker)

        # schema encoding, eg: hedwig.automatic.com/schema#/schemas/trip.created/1.0
        schema_fmt = f'{self.schema_root}#/schemas/{{message_type}}/{{message_version}}'
        schema_re = re.compile(r'([^/]+)/([^/]+)$')

        super().__init__(
            schema_fmt,
            schema_re,
            StrictVersion('1.0'),
        )

    @cached_property
    def schema_root(self) -> str:
        return self.schema['id']

    def _extract_data_helper(
        self, message_payload: Union[str, bytes], attributes: dict, use_transport_message_attributes: bool
    ) -> Tuple[MetaAttributes, dict]:
        assert isinstance(message_payload, str)

        try:
            payload = json.loads(message_payload)
        except ValueError:
            raise ValidationError('not a valid JSON')

        if not use_transport_message_attributes:
            errors = list(self._container_validator.iter_errors(payload))
            if errors:
                raise ValidationError(errors)

            data = payload['data']
            meta_attrs = MetaAttributes(
                payload['metadata']['timestamp'],
                payload['metadata']['publisher'],
                payload['metadata']['headers'],
                payload['id'],
                payload['schema'],
                payload['format_version'],
            )
        else:
            data = payload
            meta_attrs = self._decode_meta_attributes(attributes)
            if meta_attrs.format_version != self._current_format_version:
                raise ValidationError(f"Invalid format version: {meta_attrs.format_version}")
        return meta_attrs, data

    def _extract_data(
        self, message_payload: Union[str, bytes], attributes: dict, use_transport_attributes: bool
    ) -> Tuple[MetaAttributes, dict]:
        return self._extract_data_helper(
            message_payload,
            attributes,
            use_transport_message_attributes=use_transport_attributes,
        )

    def _extract_data_firehose(self, line: str) -> Tuple[MetaAttributes, dict]:
        return self._extract_data_helper(line, {}, use_transport_message_attributes=False)

    @lru_cache(maxsize=20)
    def _schema(self, message_type: str, major_version: int) -> dict:
        schema_ptr = self._schema_fmt.format(message_type=message_type, message_version=f"{major_version}.*")

        try:
            _, schema = self._validator.resolver.resolve(schema_ptr)
        except RefResolutionError:
            raise ValidationError(f'Definition not found in schema: {schema_ptr}')
        return schema

    def _verify_known_minor_version(self, message_type: str, full_version: StrictVersion):
        schema = self._schema(message_type, full_version.version[0])
        schema_full_version = StrictVersion(schema["x-version"])
        if schema_full_version.version[1] < full_version.version[1]:
            raise ValidationError(
                f'Unknown minor version: {full_version.version[1]}, last known minor version: '
                f'{schema_full_version.version[1]}'
            )

    def _decode_data(
        self,
        meta_attrs: MetaAttributes,
        message_type: str,
        full_version: StrictVersion,
        data: dict,
    ) -> dict:
        if not meta_attrs.schema.startswith(self.schema_root):
            raise ValidationError(f'message schema must start with "{self.schema_root}"')

        schema = self._schema(message_type, full_version.version[0])
        errors = list(self._validator.iter_errors(data, schema))
        if errors:
            raise ValidationError(errors)
        return data

    def _encode_data(self, data: dict) -> dict:
        assert isinstance(data, dict)
        # will get encoded in _encode_payload
        return data

    def _encode_payload_helper(
        self,
        meta_attrs: MetaAttributes,
        data: dict,
        use_transport_message_attributes: bool,
    ) -> Tuple[str, dict]:
        if not use_transport_message_attributes:
            payload = {
                'format_version': str(self._current_format_version),
                'schema': meta_attrs.schema,
                'id': meta_attrs.id,
                'metadata': {
                    'timestamp': meta_attrs.timestamp,
                    'publisher': meta_attrs.publisher,
                    'headers': meta_attrs.headers,
                },
                'data': data,
            }
            msg_attrs = deepcopy(meta_attrs.headers)
        else:
            payload = data
            msg_attrs = self._encode_meta_attributes(meta_attrs)
        return (
            json.dumps(payload, default=_json_default, allow_nan=False, separators=(',', ':'), indent=None),
            msg_attrs,
        )

    def _encode_payload(
        self, meta_attrs: MetaAttributes, data: dict, use_transport_attributes: bool
    ) -> Tuple[str, dict]:
        return self._encode_payload_helper(meta_attrs, data, use_transport_attributes)

    def _encode_payload_firehose(
        self, message_type: str, version: StrictVersion, meta_attrs: MetaAttributes, data: dict
    ) -> str:
        return self._encode_payload_helper(meta_attrs, data, use_transport_message_attributes=False)[0]

    @classmethod
    def _check_schema(cls, schema: dict) -> None:
        msg_types_found = {k: False for k in funcy.chain(settings.HEDWIG_MESSAGE_ROUTING, settings.HEDWIG_CALLBACKS)}
        # custom validation for Hedwig - TODO ideally should just be represented in json-schema file as meta schema,
        # however jsonschema lib doesn't support draft 06 which is what's really required here
        errors = []
        if not schema.get('schemas'):
            errors.append("Invalid schema file: expected key 'schemas' with non-empty value")
        else:
            for msg_type, versions in schema['schemas'].items():
                if not isinstance(versions, dict) or not versions:
                    errors.append(f"Invalid definition for: '{msg_type}', value must contain a dict of valid versions")
                    continue
                for version_pattern, definition in versions.items():
                    m = cls._version_pattern_re.match(version_pattern)
                    major_version: Optional[int]
                    if not m:
                        errors.append(f"Invalid version '{version_pattern}' for: '{msg_type}'")
                        major_version = None
                    else:
                        major_version = int(m.group(1))
                    if (msg_type, version_pattern) in msg_types_found:
                        msg_types_found[(msg_type, version_pattern)] = True
                    if not isinstance(definition, dict) and not definition:
                        errors.append(f"Invalid schema for: '{msg_type}' '{version_pattern}'")
                        continue
                    if 'x-version' not in definition:
                        errors.append(f"Invalid schema for: '{msg_type}' '{version_pattern}': missing x-version")
                        continue
                    try:
                        full_version = StrictVersion(definition['x-version'])
                        if major_version and full_version.version[0] != major_version:
                            errors.append(
                                f"Invalid full version: '{full_version}' for: '{msg_type}' '{version_pattern}'"
                            )
                    except ValueError:
                        errors.append(
                            f"Invalid full version: '{definition['x-version']}' for: '{msg_type}' "
                            f"'{version_pattern}'"
                        )

        for (msg_type, version_pattern), found in msg_types_found.items():
            if not found:
                errors.append(f"Schema not found for '{msg_type}' v{version_pattern}")

        if errors:
            raise SchemaError(str(errors))

    @staticmethod
    @FormatChecker.cls_checks('human-uuid')
    def _check_human_uuid(instance):
        if not isinstance(instance, str):
            return False
        return bool(JSONSchemaValidator._human_uuid_re.match(instance))



import dataclasses
import logging
import threading
from concurrent.futures import Future
from contextlib import ExitStack, contextmanager
from datetime import datetime
from queue import Empty, Queue
from time import time
from typing import Dict, Generator, List, Optional, Union, cast
from unittest import mock

from google.api_core.exceptions import DeadlineExceeded
from google.auth import default as google_auth_default
from google.auth import environment_vars as google_env_vars
from google.cloud import pubsub_v1
from google.cloud.pubsub_v1.subscriber.message import Message as SubscriberMessage
from google.cloud.pubsub_v1.subscriber.scheduler import Scheduler
from google.cloud.pubsub_v1.types import FlowControl, PubsubMessage, ReceivedMessage
from google.protobuf.timestamp_pb2 import Timestamp

from hedwig.backends.base import HedwigConsumerBaseBackend, HedwigPublisherBaseBackend
from hedwig.backends.utils import override_env
from hedwig.conf import settings
from hedwig.models import Message
from hedwig.utils import log

# the default visibility timeout
# ideally find by calling PubSub REST API
DEFAULT_VISIBILITY_TIMEOUT_S = 20


@contextmanager
def _seed_credentials() -> Generator[None, None, None]:
    """
    Seed environment with explicitly set credentials. Normally we'd stay away from mucking with environment vars,
    however the logic to decode `GOOGLE_APPLICATION_CREDENTIALS` isn't simple, so we let Google libraries handle it.
    """
    with ExitStack() as stack:
        if settings.GOOGLE_APPLICATION_CREDENTIALS:
            stack.enter_context(override_env(google_env_vars.CREDENTIALS, settings.GOOGLE_APPLICATION_CREDENTIALS))

        if settings.GOOGLE_CLOUD_PROJECT:
            stack.enter_context(override_env(google_env_vars.PROJECT, settings.GOOGLE_CLOUD_PROJECT))

        yield


def _auto_discover_project() -> None:
    """
    Auto discover Google project id from credentials. If project id is explicitly set, just use that.
    """
    if not settings.GOOGLE_CLOUD_PROJECT:
        # discover project from credentials
        # there's no way to get this from the Client objects, so we reload the credentials
        _, project = google_auth_default()
        assert project, "couldn't discover project"
        setattr(settings, 'GOOGLE_CLOUD_PROJECT', project)


def get_google_cloud_project() -> str:
    if not settings.GOOGLE_CLOUD_PROJECT:
        with _seed_credentials():
            _auto_discover_project()
    return settings.GOOGLE_CLOUD_PROJECT


@dataclasses.dataclass(frozen=True)
class GoogleMetadata:
    """
    Google Pub/Sub specific metadata for a Message
    """

    ack_id: str
    """
    The ID used to ack the message
    """

    subscription_path: str
    """
    Path of the Pub/Sub subscription from which this message was pulled
    """

    publish_time: datetime
    """
    Time this message was originally published to Pub/Sub
    """

    delivery_attempt: int
    """
    The delivery attempt counter received from Pub/Sub.
    The first delivery of a given message will have this value as 1. The value
    is calculated as best effort and is approximate.
    """


class GooglePubSubAsyncPublisherBackend(HedwigPublisherBaseBackend):
    def __init__(self) -> None:
        self._publisher = None

    @property
    def publisher(self):
        if self._publisher is None:
            with _seed_credentials():
                self._publisher = pubsub_v1.PublisherClient(batch_settings=settings.HEDWIG_PUBLISHER_GCP_BATCH_SETTINGS)
        return self._publisher

    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Dict[str, str]) -> Union[str, Future]:
        """
        Publishes to a Google Pub/Sub topic and returns a future that represents the publish API call. These API calls
        are batched for better performance.

        Note: despite the signature this doesn't return an actual instance of Future class, but an object that conforms
        to Future class. There's no generic type to represent future objects though.
        """
        attrs = dict((str(key), str(value)) for key, value in attrs.items())
        return self.publisher.publish(topic_path, data=data, **attrs)

    def _get_topic_path(self, message: Message) -> str:
        topic = self.topic(message)
        if isinstance(topic, tuple):
            topic, project = topic
        else:
            project = get_google_cloud_project()
        return self.publisher.topic_path(project, f'hedwig-{topic}')

    def _mock_queue_message(self, message: Message) -> "MessageWrapper":
        payload, attributes = message.serialize()
        # Pub/Sub requires bytes
        if isinstance(payload, str):
            payload = payload.encode('utf8')
            attributes['hedwig_encoding'] = 'utf8'
        publish_time = Timestamp()
        publish_time.GetCurrentTime()
        pubsub_message = PubsubMessage(
            data=payload,
            attributes=attributes,
            message_id=str(int(time() * 1000)),
            publish_time=publish_time,
        )
        # SubscriberMessage requires raw proto class, not proto-plus
        subscriber_message = SubscriberMessage(PubsubMessage.pb(pubsub_message), 'test-receipt', 1, mock.MagicMock())
        gcp_message = MessageWrapper(subscriber_message, 'test-subscription')
        return gcp_message

    def _publish(self, message: Message, payload: Union[str, bytes], attributes: Dict[str, str]) -> Union[str, Future]:
        topic_path = self._get_topic_path(message)
        # Pub/Sub requires bytes
        if isinstance(payload, str):
            payload = payload.encode('utf8')
            attributes['hedwig_encoding'] = 'utf8'
        return self.publish_to_topic(topic_path, payload, attributes)


class GooglePubSubPublisherBackend(GooglePubSubAsyncPublisherBackend):
    def publish_to_topic(self, topic_path: str, data: bytes, attrs: Dict[str, str]) -> Union[str, Future]:
        return cast(Future, super().publish_to_topic(topic_path, data, attrs)).result()


class MessageWrapper:
    def __init__(self, message: SubscriberMessage, subscription_path: str):
        self._message = message
        self._subscription_path = subscription_path

    @property
    def message(self) -> SubscriberMessage:
        return self._message

    @property
    def subscription_path(self) -> str:
        return self._subscription_path


class PubSubMessageScheduler(Scheduler):
    """
    A scheduler to use with streaming pull that simply queues all messages for the main thread to pick them up.
    """

    def __init__(self, work_queue: Queue, subscription_path: str):
        self._queue: Queue = Queue()
        self._work_queue: Queue = work_queue
        self._subscription_path: str = subscription_path

    @property
    def queue(self) -> Queue:
        """Queue: A thread-safe queue used for communication between callbacks
        and the scheduling thread."""
        return self._queue

    def schedule(self, callback, message: SubscriberMessage, *args, **kwargs) -> None:
        # callback is unused since we never set it in pull_messages
        self._work_queue.put(MessageWrapper(message, self._subscription_path))

    def shutdown(self, await_msg_callbacks=False) -> None:
        """Shuts down the scheduler and immediately end all pending callbacks."""
        # ideally we'd nack the messages in work queue, but that might take some time to finish.
        # instead, it's faster to actually process all the messages


class GooglePubSubConsumerBackend(HedwigConsumerBaseBackend):
    def __init__(self, dlq=False) -> None:
        super().__init__()
        self._subscriber: pubsub_v1.SubscriberClient = None
        self._publisher: pubsub_v1.PublisherClient = None

        if not settings.HEDWIG_SYNC:
            cloud_project = get_google_cloud_project()

            self._subscription_paths: List[str] = []
            if dlq:
                self._subscription_paths = [
                    pubsub_v1.SubscriberClient.subscription_path(cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}-dlq')
                ]
            else:
                self._subscription_paths = []
                for subscription in settings.HEDWIG_SUBSCRIPTIONS:
                    # all subscriptions live in an app's project, but cross-project subscriptions are named differently
                    if isinstance(subscription, str):
                        subscription_name = f'hedwig-{settings.HEDWIG_QUEUE}-{subscription}'
                    else:
                        subscription_name = f'hedwig-{settings.HEDWIG_QUEUE}-{subscription[1]}-{subscription[0]}'
                    self._subscription_paths.append(
                        pubsub_v1.SubscriberClient.subscription_path(cloud_project, subscription_name)
                    )
                # main queue for DLQ re-queued messages
                self._subscription_paths.append(
                    pubsub_v1.SubscriberClient.subscription_path(cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}')
                )
            self._dlq_topic_path: str = pubsub_v1.PublisherClient.topic_path(
                cloud_project, f'hedwig-{settings.HEDWIG_QUEUE}-dlq'
            )

    @property
    def subscriber(self):
        if self._subscriber is None:
            with _seed_credentials():
                self._subscriber = pubsub_v1.SubscriberClient()
        return self._subscriber

    @property
    def publisher(self):
        if self._publisher is None:
            with _seed_credentials():
                self._publisher = pubsub_v1.PublisherClient()
        return self._publisher

    def pull_messages(
        self,
        num_messages: int = 10,
        visibility_timeout: Optional[int] = None,
        shutdown_event: Optional[threading.Event] = None,
    ) -> Union[Generator, List]:
        """
        Pulls messages from PubSub subscriptions, using streaming pull, limiting to num_messages messages at a time
        """
        assert self._subscription_paths, "no subscriptions path: ensure HEDWIG_SUBSCRIPTIONS is set"

        if not shutdown_event:
            shutdown_event = threading.Event()  # pragma: no cover

        work_queue: Queue = Queue()
        futures: List[Future] = []
        flow_control: FlowControl = FlowControl(
            max_messages=num_messages,
            max_duration_per_lease_extension=visibility_timeout or DEFAULT_VISIBILITY_TIMEOUT_S,
        )

        for subscription_path in self._subscription_paths:
            # need a separate scheduler per subscription since the queue is tied to subscription path
            scheduler: PubSubMessageScheduler = PubSubMessageScheduler(work_queue, subscription_path)
            futures.append(
                self.subscriber.subscribe(
                    subscription_path, callback=None, flow_control=flow_control, scheduler=scheduler
                )
            )

        while not shutdown_event.is_set():
            try:
                message = work_queue.get(timeout=1)
                yield message
            except Empty:
                self._perform_error_counter_inactivity_reset()
                self._call_heartbeat_hook()

        for future in futures:
            future.cancel()

        # drain the queue
        try:
            while True:
                yield work_queue.get(block=False)
        except Empty:
            self._perform_error_counter_inactivity_reset()
            self._call_heartbeat_hook()

    def process_message(self, queue_message: MessageWrapper) -> None:
        # body is always bytes
        message_payload = queue_message.message.data
        attributes = queue_message.message.attributes
        if attributes.get("hedwig_encoding") == "utf8":
            message_payload = message_payload.decode('utf8')
        self.message_handler(
            message_payload,
            attributes,
            GoogleMetadata(
                queue_message.message.ack_id,
                queue_message.subscription_path,
                queue_message.message.publish_time,
                queue_message.message.delivery_attempt,
            ),
        )

    def ack_message(self, queue_message: MessageWrapper) -> None:
        queue_message.message.ack()

    def nack_message(self, queue_message: MessageWrapper) -> None:
        log(__name__, logging.INFO, "nacking message")
        queue_message.message.nack()

    @staticmethod
    def pre_process_hook_kwargs(queue_message: MessageWrapper) -> dict:
        return {'google_pubsub_message': queue_message.message}

    @staticmethod
    def post_process_hook_kwargs(queue_message: MessageWrapper) -> dict:
        return {'google_pubsub_message': queue_message.message}

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata: GoogleMetadata) -> None:
        """
        Extends visibility timeout of a message on a given priority queue for long running tasks.
        """
        if visibility_timeout_s < 0 or visibility_timeout_s > 600:
            raise ValueError("Invalid visibility_timeout_s")
        self.subscriber.modify_ack_deadline(
            subscription=metadata.subscription_path,
            ack_ids=[metadata.ack_id],
            ack_deadline_seconds=visibility_timeout_s,
        )
        self._call_heartbeat_hook(force=True)

    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: Optional[int] = None) -> None:
        """
        Re-queues everything in the Hedwig DLQ back into the Hedwig queue.

        :param num_messages: Maximum number of messages to fetch in one call. Defaults to 10.
        :param visibility_timeout: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
        """
        topic_path = pubsub_v1.PublisherClient.topic_path(get_google_cloud_project(), f'hedwig-{settings.HEDWIG_QUEUE}')
        assert len(self._subscription_paths) == 1, "multiple subscriptions found"
        subscription_path = self._subscription_paths[0]

        log(__name__, logging.INFO, "Re-queueing messages from {} to {}".format(subscription_path, topic_path))
        while True:
            try:
                queue_messages: List[ReceivedMessage] = self.subscriber.pull(
                    subscription=subscription_path,
                    max_messages=num_messages,
                    retry=None,
                    timeout=settings.GOOGLE_PUBSUB_READ_TIMEOUT_S,
                ).received_messages
            except DeadlineExceeded:
                break

            if not queue_messages:
                break

            log(__name__, logging.INFO, "got {} messages from dlq".format(len(queue_messages)))
            for queue_message in queue_messages:
                try:
                    if visibility_timeout:
                        self.subscriber.modify_ack_deadline(
                            subscription=subscription_path,
                            ack_ids=[queue_message.ack_id],
                            ack_deadline_seconds=visibility_timeout,
                        )

                    future = self.publisher.publish(
                        topic_path, data=queue_message.message.data, **queue_message.message.attributes
                    )
                    # wait for success
                    future.result()
                    log(
                        __name__,
                        logging.DEBUG,
                        'Re-queued message from DLQ {} to {}'.format(subscription_path, topic_path),
                        extra={'message_id': queue_message.message.message_id},
                    )

                    self.subscriber.acknowledge(subscription=subscription_path, ack_ids=[queue_message.ack_id])
                except Exception:
                    log(
                        __name__,
                        logging.ERROR,
                        'Exception in requeue message from {} to {}'.format(subscription_path, topic_path),
                        exc_info=True,
                    )

            log(__name__, logging.INFO, "Re-queued {} messages".format(len(queue_messages)))



import abc
from collections import namedtuple
from distutils.version import StrictVersion
from typing import Any, Tuple, Union, Dict, Pattern

from hedwig.conf import settings
from hedwig.exceptions import ValidationError
from hedwig.models import Message, Metadata

MetaAttributes = namedtuple('MetaAttributes', ['timestamp', 'publisher', 'headers', 'id', 'schema', 'format_version'])


class HedwigBaseValidator:
    """
    Base class responsible for serializing / encoding and deserializing / decoding messages into / from format on the
    wire.
    """

    _schema_re: Pattern
    """
    A regex that matches encoded schema and matches 2 groups: message_type, message_version
    """

    _schema_fmt: str
    """
    A f-string that is used to encode schema that contains two placeholders: message_type, message_version
    """

    _current_format_version: StrictVersion

    def __init__(self, schema_fmt: str, schema_re: Pattern, current_format_version: StrictVersion):
        self._schema_fmt = schema_fmt
        self._schema_re = schema_re
        self._current_format_version = current_format_version

    @abc.abstractmethod
    def _extract_data(
        self, message_payload: Union[str, bytes], attributes: dict, use_transport_attributes: bool
    ) -> Tuple[MetaAttributes, Any]:
        """
        Extracts data from the on-the-wire payload
        """

    @abc.abstractmethod
    def _extract_data_firehose(self, line: str) -> Tuple[MetaAttributes, Any]:
        """
        Extracts data from firehose line
        """

    @abc.abstractmethod
    def _decode_data(
        self,
        meta_attrs: MetaAttributes,
        message_type: str,
        full_version: StrictVersion,
        data: Any,
    ) -> Any:
        """
        Validates decoded data
        """

    def _encode_message_type(self, message_type: str, version: StrictVersion) -> str:
        """
        Encodes message type in outgoing message attribute
        """
        return self._schema_fmt.format(message_type=message_type, message_version=version)

    def _decode_message_type(self, schema: str) -> Tuple[str, StrictVersion]:
        """
        Decode message type from meta attributes
        """
        try:
            m = self._schema_re.search(schema)
            if m is None:
                raise ValueError
            schema_groups = m.groups()
            message_type = schema_groups[0]
            full_version = StrictVersion(schema_groups[1])
        except (AttributeError, ValueError):
            raise ValidationError(f'Invalid schema found: {schema}')
        return message_type, full_version

    def _verify_known_minor_version(self, message_type: str, full_version: StrictVersion):
        """
        Validate that minor version is known
        """
        raise NotImplementedError

    def _verify_headers(self, headers: dict):
        """
        Validate headers are sane
        """
        for k, v in headers.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise ValidationError(f"Invalid header key or value: '{k}': '{v}' - must be strings")
            if k.startswith("hedwig_"):
                raise ValidationError(f"Invalid header key: '{k}' - can't begin with reserved namespace 'hedwig_'")

    def _deserialize(
        self,
        message_payload: Union[str, bytes],
        attributes: dict,
        provider_metadata: Any,
        use_transport_attributes: bool,
    ) -> Message:
        """
        Deserialize a message from the on-the-wire format
        :param message_payload: Raw message payload as received from the backend
        :param provider_metadata: Provider specific metadata
        :param attributes: Message attributes from the transport backend
        :returns: Message object if valid
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        meta_attrs, extracted_data = self._extract_data(message_payload, attributes, use_transport_attributes)
        self._verify_headers(meta_attrs.headers)
        message_type, version = self._decode_message_type(meta_attrs.schema)
        data = self._decode_data(meta_attrs, message_type, version, extracted_data)

        return Message(
            id=meta_attrs.id,
            metadata=Metadata(
                timestamp=meta_attrs.timestamp,
                headers=meta_attrs.headers,
                publisher=meta_attrs.publisher,
                provider_metadata=provider_metadata,
            ),
            data=data,
            type=message_type,
            version=version,
        )

    def deserialize(self, message_payload: Union[str, bytes], attributes: dict, provider_metadata: Any) -> Message:
        """
        Deserialize a message from the on-the-wire format
        :param message_payload: Raw message payload as received from the backend
        :param provider_metadata: Provider specific metadata
        :param attributes: Message attributes from the transport backend
        :returns: Message object if valid
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        return self._deserialize(
            message_payload, attributes, provider_metadata, settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES
        )

    def deserialize_containerized(self, message_payload: Union[str, bytes]) -> Message:
        """
        Deserialize a message assuming containerized format regardless of configured setting.
        :param message_payload: Raw message payload as received from the backend
        :returns: Message object if valid
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        return self._deserialize(message_payload, {}, None, False)

    def deserialize_firehose(self, line: str) -> Message:
        """
        Deserialize a message from a line in firehose file. This is slightly different from on-the-wire format:

        1. It always uses container format (i.e. HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES is ignored)
        2. For binary file formats, its possible data is encoded as binary base64 blob rather than JSON.
        3. Known minor versions isn't verified - knowing major version schema is good enough to read firehose.

        :param line: Raw line read from firehose file
        :returns: Message object if valid
        :raise: :class:`hedwig.ValidationError` if validation fails.
        """
        meta_attrs, extracted_data = self._extract_data_firehose(line)
        message_type, version = self._decode_message_type(meta_attrs.schema)
        data = self._decode_data(meta_attrs, message_type, version, extracted_data)

        return Message(
            id=meta_attrs.id,
            metadata=Metadata(
                timestamp=meta_attrs.timestamp,
                headers=meta_attrs.headers,
                publisher=meta_attrs.publisher,
                provider_metadata=None,
            ),
            data=data,
            type=message_type,
            version=version,
        )

    def _encode_payload(
        self, meta_attrs: MetaAttributes, data: Any, use_transport_attributes: bool
    ) -> Tuple[Union[str, bytes], dict]:
        """
        Encodes on-the-wire payload
        """
        raise NotImplementedError

    def _encode_payload_firehose(
        self, message_type: str, version: StrictVersion, meta_attrs: MetaAttributes, data: Any
    ) -> str:
        """
        Encodes firehose line
        """
        raise NotImplementedError

    def _serialize(self, message: Message, use_transport_attributes: bool) -> Tuple[Union[str, bytes], dict]:
        self._verify_known_minor_version(message.type, message.version)
        self._verify_headers(message.headers)
        schema = self._encode_message_type(message.type, message.version)
        meta_attrs = MetaAttributes(
            message.timestamp,
            message.publisher,
            message.headers,
            message.id,
            schema,
            self._current_format_version,
        )
        message_payload, msg_attrs = self._encode_payload(meta_attrs, message.data, use_transport_attributes)
        # validate payload from scratch before publishing
        self._deserialize(message_payload, msg_attrs, None, use_transport_attributes)
        return message_payload, msg_attrs

    def serialize(self, message: Message) -> Tuple[Union[str, bytes], dict]:
        """
        Serialize a message for appropriate on-the-wire format
        :return: Tuple of message payload and transport attributes
        """
        return self._serialize(message, settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES)

    def serialize_containerized(self, message: Message) -> Union[str, bytes]:
        """
        Serialize a message using containerized format regardless of configured settings. In most cases, you just want
        to use `.serialize`.
        :return: Message payload
        """
        return self._serialize(message, False)[0]

    def serialize_firehose(self, message: Message) -> str:
        """
        Serialize a message for appropriate firehose file format. See
        :meth:`hedwig.validators.base.HedwigBaseValidator.deserialize_firehose` for details.
        :return: Tuple of message payload and transport attributes
        """
        schema = self._encode_message_type(message.type, message.version)
        meta_attrs = MetaAttributes(
            message.timestamp,
            message.publisher,
            message.headers,
            message.id,
            schema,
            self._current_format_version,
        )
        message_payload = self._encode_payload_firehose(message.type, message.version, meta_attrs, message.data)
        # validate payload from scratch
        self.deserialize_firehose(message_payload)
        return message_payload

    def _decode_meta_attributes(self, attributes: Dict[str, str]) -> MetaAttributes:
        """
        Decodes meta attributes from transport attributes
        :param attributes: Message attributes from the transport backend
        :return:
        """
        assert settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES

        for attr in (
            'hedwig_format_version',
            'hedwig_id',
            'hedwig_message_timestamp',
            'hedwig_publisher',
            'hedwig_schema',
        ):
            value = attributes.get(attr)
            if not isinstance(value, str):
                raise ValidationError(f"Invalid message attribute: {attr} must be string, found: {value}")

        headers = {k: v for k, v in attributes.items() if not k.startswith("hedwig_")}
        return MetaAttributes(
            int(attributes['hedwig_message_timestamp']),
            attributes['hedwig_publisher'],
            headers,
            attributes['hedwig_id'],
            attributes['hedwig_schema'],
            StrictVersion(attributes['hedwig_format_version']),
        )

    def _encode_meta_attributes(self, meta_attrs: MetaAttributes) -> Dict[str, str]:
        """
        Encodes meta attributes as transport attributes
        :param meta_attrs:
        :return:
        """
        assert settings.HEDWIG_USE_TRANSPORT_MESSAGE_ATTRIBUTES

        attrs = {
            'hedwig_format_version': str(meta_attrs.format_version),
            'hedwig_id': str(meta_attrs.id),
            'hedwig_message_timestamp': str(meta_attrs.timestamp),
            'hedwig_publisher': meta_attrs.publisher,
            'hedwig_schema': meta_attrs.schema,
        }
        attrs.update(meta_attrs.headers)
        return attrs



from distutils.version import StrictVersion

from opentelemetry.version import __version__

VERSION_1_0_0 = StrictVersion("1.0.0")

if StrictVersion(__version__) >= VERSION_1_0_0:
    from opentelemetry.propagate import extract as extract_10, inject as inject_10  # type: ignore
    from opentelemetry.propagators.textmap import DefaultGetter as Getter  # type: ignore
    from opentelemetry.trace.span import (
        format_trace_id as get_hexadecimal_trace_id,
        format_span_id as get_hexadecimal_span_id,
    )
    from opentelemetry.propagate import set_global_textmap

    id_generator_class = "opentelemetry.sdk.trace.RandomIdGenerator"

    def get_traceparent_string(trace_id, span_id, trace_flags):
        return "00-{trace_id}-{span_id}-{:02x}".format(
            trace_flags,
            trace_id=get_hexadecimal_trace_id(trace_id),
            span_id=get_hexadecimal_span_id(span_id),
        )

else:
    from opentelemetry.propagators import extract, inject  # type: ignore
    from opentelemetry.trace.propagation.textmap import DictGetter as Getter  # type: ignore
    from opentelemetry.trace.span import get_hexadecimal_trace_id, get_hexadecimal_span_id
    from opentelemetry.propagators import set_global_textmap

    id_generator_class = "opentelemetry.sdk.trace.RandomIdsGenerator"

    def get_traceparent_string(trace_id, span_id, trace_flags):
        return "00-{:032x}-{:016x}-{:02x}".format(
            trace_id,
            span_id,
            trace_flags,
        )


if StrictVersion(__version__) >= VERSION_1_0_0:

    def extract(getter, carrier):  # type: ignore
        return extract_10(carrier)

    def inject(setter, carrier):  # type: ignore
        return inject_10(carrier)


__ALL__ = [
    "Getter",
    "extract",
    "get_hexadecimal_span_id",
    "get_hexadecimal_trace_id",
    "id_generator_class",
    "inject",
    "set_global_textmap",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hedwig/protobuf/options.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='hedwig/protobuf/options.proto',
  package='hedwig',
  syntax='proto2',
  serialized_options=b'\n\034io.github.cloudchacho.hedwigZ)github.com/cloudchacho/hedwig-go/protobuf',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dhedwig/protobuf/options.proto\x12\x06hedwig\x1a google/protobuf/descriptor.proto\"T\n\x0eMessageOptions\x12\x15\n\rmajor_version\x18\x01 \x01(\r\x12\x15\n\rminor_version\x18\x02 \x01(\r\x12\x14\n\x0cmessage_type\x18\x03 \x01(\t:R\n\x0fmessage_options\x12\x1f.google.protobuf.MessageOptions\x18\xa3\x90\x03 \x01(\x0b\x32\x16.hedwig.MessageOptionsBI\n\x1cio.github.cloudchacho.hedwigZ)github.com/cloudchacho/hedwig-go/protobuf'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])


MESSAGE_OPTIONS_FIELD_NUMBER = 51235
message_options = _descriptor.FieldDescriptor(
  name='message_options', full_name='hedwig.message_options', index=0,
  number=51235, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)


_MESSAGEOPTIONS = _descriptor.Descriptor(
  name='MessageOptions',
  full_name='hedwig.MessageOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='major_version', full_name='hedwig.MessageOptions.major_version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='minor_version', full_name='hedwig.MessageOptions.minor_version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_type', full_name='hedwig.MessageOptions.message_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=159,
)

DESCRIPTOR.message_types_by_name['MessageOptions'] = _MESSAGEOPTIONS
DESCRIPTOR.extensions_by_name['message_options'] = message_options
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MessageOptions = _reflection.GeneratedProtocolMessageType('MessageOptions', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEOPTIONS,
  '__module__' : 'hedwig.protobuf.options_pb2'
  # @@protoc_insertion_point(class_scope:hedwig.MessageOptions)
  })
_sym_db.RegisterMessage(MessageOptions)

message_options.message_type = _MESSAGEOPTIONS
google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(message_options)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from distutils.version import StrictVersion
import random
from unittest import mock

import pytest

from hedwig.exceptions import ValidationError, CallbackNotFound
from hedwig.models import Message

from tests.models import MessageType


class TestMessageMethods:
    publisher = 'myapi'

    def test_new(self, message_data):
        message = Message.new(
            MessageType.trip_created,
            StrictVersion('1.0'),
            message_data['data'],
            message_data['id'],
            message_data['metadata']['headers'],
        )

        assert message.id == message_data['id']
        assert message.version == StrictVersion('1.0')
        assert message.headers == message_data['metadata']['headers']
        assert message.type == 'trip_created'
        assert message.data == message_data['data']

    @mock.patch('hedwig.callback.Callback.find_by_message', side_effect=CallbackNotFound)
    def test_validate_missing_task(self, _, message):
        with pytest.raises(ValidationError):
            _ = message.callback

    @mock.patch('tests.handlers._trip_created_handler', autospec=True)
    def test_exec_callback(self, mock_trip_created_handler, message):
        message.exec_callback()
        mock_trip_created_handler.assert_called_once_with(message)

    @mock.patch('hedwig.models.get_consumer_backend')
    def test_extend_visibility_timeout(self, mock_get_consumer_backend, message):
        visibility_timeout_s = random.randint(0, 1000)

        message.extend_visibility_timeout(visibility_timeout_s)

        mock_get_consumer_backend.assert_called_once_with()

        mock_get_consumer_backend.return_value.extend_visibility_timeout.assert_called_once_with(
            visibility_timeout_s, message.provider_metadata
        )

    def test_getter_timestamp(self, message):
        assert message.timestamp == message.metadata.timestamp

    def test_getter_publisher(self, message):
        assert message.publisher == message.metadata.publisher
